#include <fstream>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <iostream>
#include <vector>
#include "loadPLY.h"

using namespace std;

int loadPLY ( char *fileName, vector< vertexType > &vVec, vector< indexType > &tVec )
{
  int retCode = 0;
  int i = 0, vi = 0, fi = 0;
  int vertexCount = 0, faceCount = 0, vertexPCount = 0, facePCount = 0;
  ifstream in(fileName);
  char token[256], element[256];
  enum State { READ_HEADER, READ_VERTEX, READ_FACE, READ_TOKEN };
  State state = READ_HEADER;
  vertexType vertex;
  indexType triangle;

  if (in) {
    while (in.eof() == false) {
      switch (state) {
	case READ_HEADER:
	  in >> token;
	  if (!strcmp(token,"element")) {
	    in >> element;
	    if (!strcmp(element,"vertex")) 
	      in >> vertexCount;
	    else
	      if (!strcmp(element,"face"))
		in >> faceCount;
	      else
		cerr << "ERROR: Unknown element type in loadPLY" << endl;
	  }
	  else {
	    if (!strcmp(token,"property")) {
	      switch (*element) {
		case 'v':
		  vertexPCount++;
		  break;
		case 'f':
		  facePCount++;
		  break;
		default:
		  cerr << "ERROR: Invalid Property" << endl;
	      }
	      in >> token >> token;
	    }
	    else
	      if (!strcmp(token,"end_header")) {
		state = READ_VERTEX;
		vi = fi = 0;
		cerr << "cout (PLY: FC:" 
		  << faceCount << ", VC: " << vertexCount << ")" 
		  << flush;
	      }
	  }
	  break;
	case READ_VERTEX:
	  in >> vertex[0] >> vertex[1] >> vertex[2];
	  vVec.push_back(vertex);
	  for (i = 3; i < vertexPCount; i++)
	    in >> token;
	  if (++vi == vertexCount)
	    state = READ_FACE;
	  break;
	case READ_FACE:
	  in >> vi;
	  if (vi != 3) {
	    cerr << "Bad face type. Triangulate first." << endl;
	    return 0;
	  };
	  in >> triangle[0] >> triangle[1] >> triangle[2];
	  tVec.push_back(triangle);
	  if ((++fi) == faceCount)
	    state = READ_TOKEN;
	  break;
	case READ_TOKEN:
	  retCode = 1;
	  in >> token;
	  break;
      }
    }
  }
  cerr << "done. " << endl;
  return retCode;
}

