// ----------------------------------------------
// Demo class for Fast Striping
// Author: Patrick Reuter
// date: September 12th, 2000
// ----------------------------------------------

#ifndef WIN32
#include <iostream.h>
#endif

#ifdef WIN32
#include <winsock.h>
#include <time.h>
#else
#include <sys/types.h>
#include <sys/time.h>
#endif

#include <GL/glut.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <vector>
using namespace std;

// fileIO
#include "loadPLY.h"
int fileStatus;
char* fname;
#if 0
vector< vector < float > > vertexVec;
vector< vector < int > > triVec;  
#else
vector< vertexType > vertexVec;
vector< indexType > triVec;
#endif

// Fast Striping 
#include "TriangleAdjacencyGraph.h"
TriangleAdjacencyGraph graph;
unsigned int numPrimitives;
vector < vector < Index > > primIndex;
vector<int> primitiveType;


static int drawStrips = 0;
static int colorStrips = 0;

// Global Variables
static int newModelViewMatrix = 1;
static int visible = 1;
bool newMesh=true;

// Trackball simulation
#include "trackball.h"
static float quat[4];
static float lastquat[4];
static float m[4][4];
static float scalefactor=1;
static int spinning=0, moving=0, scaling=0;
static int beginx, beginy;

// Window size
int W = 512, H = 512;

// Model's Bounding Box
static float minCorner[3], maxCorner[3];
static float translatex=0, translatey=0, translatez=0;  
static float size;
static int vertices;

// Switches for Toggle Menus
static int ccw=0;
static int cull=1;

// determine fps 
static double frameRate = 0.0;
double lastTime;
static float totalFrameRate = 0;
double getSystemTime (void) {
  struct timeval tv;
  double time;
  gettimeofday(&tv, 0);
  time = double(tv.tv_usec) / 1000000.0 + double(tv.tv_sec);
  return time;
};

// Update Benchmarks in the Window Bottom
void updateInfoLine (void) 
{
  char buffer[256];
  sprintf(buffer, "Vertices: %d      Framerate: %3.2f", vertices, frameRate);  
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glOrtho (0.0, W, 0.0, H, -1.0, 1.0);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glDisable(GL_LIGHTING);
  glDisable(GL_TEXTURE_2D); 
  glColor3f (0, 1, 0);
  glRasterPos2f(10, 10);
  for (int i = 0; buffer[i]; i++)
    glutBitmapCharacter(GLUT_BITMAP_9_BY_15, buffer[i]);
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();      
}

// Now call our fast Striping algorithm
void fastStriping(unsigned int numIterations) {

  // don't strip more than once
  if ( drawStrips == 1 )
     return;

  double t1, t2;
  int cost;

  // unsigned int numPrimitives;

  // start Benchmark
  t1 = getSystemTime();

  // reserve space (not needed, but little higher performance)
  graph.reserve (vertexVec.size(), triVec.size() );  

  for (unsigned int i = 0; i < triVec.size(); i ++) {
    // add the triangle to the data structure
    graph.addTriangle(triVec[i][0], triVec[i][1], triVec[i][2]);
  }
  // calculate the primitives 
  // (n iterations, doStrips = true, doFans = true, FanTreshold = 16)
  cost = graph.calcOptPrim (numIterations, true, true, 16 );

  numPrimitives = graph.primitiveCount();
  primitiveType.resize(numPrimitives);
  primIndex.resize(numPrimitives);
  for (unsigned int i = 0; i < numPrimitives; i++) {
    primitiveType[i] = graph.getPrimitive(primIndex[i]);
  };

  // end Benchmark
  t2 = getSystemTime();
  std::cerr << "Fast Striping in " << (t2 - t1) << ", we'll need " << cost << " vertices to be sent to the renderer. " << endl;


  // just to show the vertex count
  vertices = cost;

  // enable draw strips  
  drawStrips = 1;

}; 


// Control the context menu
void menuControl (int value)
{
  switch (value) {
    case 90:
      glutFullScreen(); 
      break;
    case 40:
      ccw = !ccw;
      glFrontFace ( ccw ? GL_CCW : GL_CW ); 
      break;
    case 42:
      cull = !cull;
      if (cull)
        glEnable(GL_DEPTH_TEST);
      else
        glDisable(GL_DEPTH_TEST);
      break;
    case 11:
      glPolygonMode(GL_FRONT_AND_BACK, GL_POINT);
      break;
    case 12:
      glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
      break;
    case 13:
      glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
      break;
    case 91:
      glClearColor(1.0, 1.0, 1.0, 0);
      break;
    case 92:
      glClearColor(0.0,0.0, 0.0, 0);
      break;
    case 100:
      exit(1);
      break;
    case 14:
      fastStriping( 1 );
      break;
    case 15:
      fastStriping( 10 );
      break;
    case 16:
      fastStriping( 100 );
      break;
    case 17:
      fastStriping( 1000 );
      break;
    case 20:
      colorStrips = !colorStrips;
      break;
  };
  newModelViewMatrix = 1;
  glutPostRedisplay();
}

// Init the context menu
void menuInit(void) 
{
  int rootMenu, stripingMenu, splatMenu, stateSettingMenu;

  stripingMenu = glutCreateMenu(menuControl);
  glutAddMenuEntry("Fast Striping!",14);
  glutAddMenuEntry("Sampling 10 Iterations",15);
  glutAddMenuEntry("Sampling 100 Iterations",16);
  glutAddMenuEntry("Sampling 1000 Iterations",17);
    
  splatMenu = glutCreateMenu(menuControl);
  glutAddMenuEntry("Points", 11);
  glutAddMenuEntry("Wireframe", 12);
  glutAddMenuEntry("Fill", 13);
  glutAddMenuEntry("Color strips", 20);

  stateSettingMenu = glutCreateMenu(menuControl);
  glutAddMenuEntry("Switch Front face", 40);
  glutAddMenuEntry("Toggle Depth Testing", 42);

  rootMenu = glutCreateMenu(menuControl);
  glutAddSubMenu("Striping ... ", stripingMenu);
  glutAddSubMenu("Triangle Drawing", splatMenu);
  glutAddSubMenu("OpenGL settings", stateSettingMenu);

  glutAddMenuEntry("White background", 91);
  glutAddMenuEntry("Black background", 92);
  glutAddMenuEntry("Full screen", 90);
  glutAddMenuEntry("Quit", 100);  

  glutAttachMenu(GLUT_RIGHT_BUTTON);
}

// init OpenGL attributes
void initGL(void) 
{
  glEnable(GL_COLOR_MATERIAL);
  glEnable(GL_DEPTH_TEST);
  glCullFace(GL_FRONT);
  glEnable(GL_CULL_FACE);
  glLineWidth(2.0);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective( 45, (GLfloat) W/(GLfloat) H , size / 100.0, size * 2.0 );
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(0.0, 0.0, size, 0.0, 0.0, 0.0, 0.0, 1.0, 0.); 
  glPushMatrix();   

  glClearColor (0.0, 0.0, 0.0, 0.0);  
  glFrontFace ( ccw ? GL_CCW : GL_CW ); 

  newMesh = false;
}


// recalculation of the model view
void recalcModelView(void)
{
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective( 45, double(W)/double(H) , size / 100.0, 
      (size/scalefactor) + (size * 1));

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(0.0, 0.0, size/scalefactor, 0.0, 0.0, 0.0, 0.0, 1.0, 0.); 

  build_rotmatrix(m, quat);
  glMultMatrixf(&m[0][0]);

  if (scalefactor == 1.0)
    glDisable(GL_NORMALIZE);
  else
    glEnable(GL_NORMALIZE);


  glTranslatef (-translatex, -translatey, -translatez);
  newModelViewMatrix = 0;
}

// Redraw the screen
void redraw(void)
{
  //unsigned int numPrimitives;
  double currentTime = getSystemTime();
  if (newModelViewMatrix)
    recalcModelView();
  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  frameRate = double(1) / double(currentTime - lastTime);
  lastTime = currentTime;
  updateInfoLine();

  glEnableClientState( GL_VERTEX_ARRAY );
  glVertexPointer( 3, GL_FLOAT, 0, &vertexVec[0] );
    
  // now draw either strips/fans or raw triangles
  glColor3f(0.5, 0.5, 0.5); // in gray color

  if (drawStrips) {

    if ( colorStrips )
        srand(0);// for random color
      
    for (unsigned int i = 0; i < numPrimitives; i++)
    {
      // create a random face color
      if ( colorStrips )
          glColor3ub( (unsigned char)( 255.0f * rand() / float( RAND_MAX ) ) ,
                      (unsigned char)( 255.0f * rand() / float( RAND_MAX ) ) ,
                      (unsigned char)( 255.0f * rand() / float( RAND_MAX ) ) );

      glDrawElements( primitiveType[i],
                      primIndex[i].size(),
                      GL_UNSIGNED_INT,
                      &primIndex[i][0] );
    }
  }
  else {
    glDrawElements( GL_TRIANGLES,
                    3 * triVec.size(),
                    GL_UNSIGNED_INT,
                    &triVec[0] );
  };
  
  glDisableClientState( GL_VERTEX_ARRAY );
  glPushAttrib(GL_ENABLE_BIT);
  glPopAttrib();

  glutSwapBuffers();
};

// The user reshaped the Window size
void reshape (int w, int h)
{
  glViewport (0, 0, (GLsizei) w, (GLsizei) h); 
  W = w;
  H = h;

  glMatrixMode (GL_PROJECTION);
  glLoadIdentity ();
  gluPerspective(45, (GLfloat) w/(GLfloat) h, size/100, size*2);
  glMatrixMode(GL_MODELVIEW);
}

// Mouse moved to change size or orientation of the model
void motion(int x, int y)
{
  if (scaling) {
    scalefactor = scalefactor * (1.0 + (((float) (beginy - y)) / H));
    beginx = x;
    beginy = y;
    newModelViewMatrix = 1;
    newMesh = 1;
    glutPostRedisplay();
    return;
  }
  if (moving) {
    trackball(lastquat,
        (2.0 * beginx - (W)) / (W),
        ((H) - 2.0 * beginy) / (H),
        (2.0 * x - (W)) / (W),
        ((H) - 2.0 * y) / (H));
    add_quats(lastquat, quat, quat);

    beginx = x;
    beginy = y;
    spinning = 1;
  }
  newModelViewMatrix = 1;
}

// Mouse button pressed
void mouse(int button, int state, int x, int y)
{
  if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) {
    spinning = 0;
    moving = 1;
    beginx = x;
    beginy = y;
    scaling = (glutGetModifiers() & GLUT_ACTIVE_SHIFT) ? 1 : 0;  
  }
  if (button == GLUT_LEFT_BUTTON && state == GLUT_UP) {
    moving = 0;
  }
}

// The user interferred via the keyboard
void keyboard (unsigned char key, int x, int y)
{
  switch (key) {
    case 27:
      exit(0);
      break;
    case 'q' : 
      exit(0);
      break;
    default:
      break;  
  } 
  updateInfoLine();
  if (newModelViewMatrix) 
    recalcModelView();
}

// Animation
void animate(void) {
  bool redisplay = false;
  if (spinning) {
    add_quats(lastquat, quat, quat);
    newModelViewMatrix = 1;
    //newMesh = 1;
    redisplay = true;
  }

  if (newModelViewMatrix) 
    recalcModelView();

  if (newMesh) {
    redisplay = true;
    totalFrameRate += frameRate;
    newMesh = false;

    updateInfoLine();
  }

  if (redisplay)
    glutPostRedisplay();
}

// Visibility
void vis(int vis)
{
  visible = (vis == GLUT_VISIBLE);
}

// Calculate Bounding Box
void boundingBox(void) {
  for (unsigned int i = 0; i < triVec.size(); i++) {
    for (int j = 0; j < 3; j++) {
      for (int t = 0; t < 3; t++) {
        if (vertexVec[triVec[i][j]][t] > maxCorner[t] || (i == 0 && j == 0 && t == 0))
          maxCorner[t] = vertexVec[triVec[i][j]][t];
        if (vertexVec[triVec[i][j]][t] < minCorner[t] || (i == 0 && j == 0 && t == 0))
          minCorner[t] = vertexVec[triVec[i][j]][t];
      }
    }
  }

  // calculate Translation  
  translatex = (maxCorner[0]+minCorner[0])/2;
  translatey = (maxCorner[1]+minCorner[1])/2;
  translatez = (maxCorner[2]+minCorner[2])/2;

  // calculate Bounding Box size
  size = sqrt ( (maxCorner[0] - minCorner[0]) * (maxCorner[0] - minCorner[0]) +
      (maxCorner[1] - minCorner[1]) * (maxCorner[1] - minCorner[1]) +
      (maxCorner[2] - minCorner[2]) * (maxCorner[2] - minCorner[2])) ;

  // just to show the vertex count
  vertices = triVec.size() * 3;
};

int main(int argc, char** argv)
{    
  glutInit(&argc, argv);
  glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH | GLUT_MULTISAMPLE);
  glutInitWindowPosition (100, 100);
  glutCreateWindow (argv[0]);
  glutReshapeWindow(W, H);  
  glutIdleFunc(animate);
  glutDisplayFunc(redraw); 
  glutReshapeFunc(reshape);
  glutVisibilityFunc(vis);
  glutKeyboardFunc(keyboard);
  glutMotionFunc(motion);
  glutMouseFunc(mouse);

  if (argc == 2) {
    fname = argv[1];
    std::cout << "Opening File ... " << std::endl;
    fileStatus = loadPLY(fname, vertexVec, triVec);
    if (!fileStatus) { 
      std::cout << "Error opening " << fname << ". Exiting" << std::endl;
      exit(1);
    }
  }
  else
  {
    std::cout << "Usage: fstripDemo <filename.ply> " << std::endl;
    std::cout << "-------------------------------- " << std::endl;
    std::cout << "--- reads ASCII ply files ------ " << std::endl << std::endl;
    exit(1);
  };  


  initGL();          // OpenGL settings
  menuInit();        // Menu settings
  boundingBox();       // Init model

  // initialize Trackball
  trackball(quat , 0.0, 0.0, 0.0, 0.0);

  glutDisplayFunc(redraw); 
  glutMainLoop();

  return 0;
}
